//******************************************************************
//* UNIVERSAL MOTOR CONTROLLER
//* LAST UPDATED 09/03/2008
//* FIRMWARE BY FRED EADY FOR SERVO MAGAZINE
//* CHANGES/ADDITIONS
//******************************************************************
#include <pic18.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <htc.h>
#include <stdlib.h>

// Settings for PIC18F2620
__CONFIG(1,IESODIS & FCMDIS & RCIO);
__CONFIG(2,BORDIS & PWRTDIS & WDTDIS);
__CONFIG(3,MCLREN & LPT1DIS & PBDIGITAL);
__CONFIG(4,DEBUGEN & LVPDIS & XINSTDIS & STVREN);
__CONFIG(5,UNPROTECT);
__CONFIG(6,WRTEN);
__IDLOC(0903);

//******************************************************************
//*	FUNCTION PROTOTYPES
//******************************************************************
void init(void);
unsigned int recvchar(void);
char CharInQueue(void);
void initEUSART(void);
void reset_mtr (char start_degrees, char revs);
void interrupt ZVC_EUSART(void);
//******************************************************************
//*	I/O DEFINITIONS
//******************************************************************
#define	triac_gate_sink_0	LATA0
#define	triac_gate_sink_1	LATA1
#define	triac_gate_sink_2	LATA2
#define	led					LATC0
#define triac_switch		RC7
//******************************************************************
//*	PIC18F2620 GLOBAL DEFINITIONS
//******************************************************************
typedef struct {
					char	int0_edge:1;		
					char	zvc:1;
					char	mtr_running;
				} FFLAGS;

FFLAGS flags;
unsigned int need4speed;
char usart_data,tmphead;
char refresh_speed_cntr;
char current_trigger_angle, new_trigger_angle;
unsigned int tmr0_trigger_val,temp16;
//******************************************************************
//*	PIC18F2620 CONSTANTS
//******************************************************************
#define REFRESH_SPEED_RATE  	6
#define	hi						1
#define lo						0
//******************************************************************
//    MACROS AND PORT/PIN DEFINITIONS
//******************************************************************
#define bitset(var, bitno) ((var) |= 1 << (bitno))
#define bitclr(var, bitno) ((var) &= ~(1 << (bitno)))
#define make8(var,offset)	((unsigned int)var >> (offset * 8)) & 0x00FF
#define	make16(varhigh,varlow)	(((unsigned int)varhigh & 0xFF)* 0x100) + ((unsigned int)varlow & 0x00FF)
#define make32(var1,var2,var3,var4) \
		((unsigned long)var1<<24)+((unsigned long)var2<<16)+ \
		((unsigned long)var3<<8)+((unsigned long)var4)		
#define make32i(var1,var2) ((unsigned long)var1<<16)+((unsigned long)var2)
//******************************************************************
//*	TRIAC GATE MACROS
//******************************************************************
#define		block_triac			triac_gate_sink_0 = 1;	\
								triac_gate_sink_1 = 1;	\
								triac_gate_sink_2 = 1;
								
#define		trigger_triac		triac_gate_sink_0 = 0;	\
								triac_gate_sink_1 = 0;	\
								triac_gate_sink_2 = 0;

//******************************************************************
//*	INTERRUPT DEFINITIONS
//******************************************************************
#define		disable_RXint	  	RCIE = 0
#define		enable_RXint	  	RCIE = 1
#define		disable_TMR2int  	TMR2IE = 0
#define		enable_TMR2int	 	TMR2IE = 1
#define		disable_TMR3int  	TMR3IE = 0
#define		enable_TMR3int	 	TMR3IE = 1
#define		disable_TMR1int  	TMR1IE = 0
#define		enable_TMR1int	 	TMR1IE = 1
#define		disable_TMR0int  	TMR0IE = 0
#define		enable_TMR0int	 	TMR0IE = 1
#define		disable_INT0int		INT0IE = 0
#define		enable_INT0int		INT0IE = 1

#define		enable_GLOBALint  	IPEN = 0;				\
							  	GIE = 1;				\
							  	PEIE = 1;
#define		disable_GLOBALint 	IPEN = 0;				\
							  	GIE = 0;				\
							  	PEIE = 0;
#define		TIMER0ON			TMR0ON = 1
#define		TIMER0OFF			TMR0ON = 0
#define		TIMER1ON			TMR1ON = 1
#define		TIMER1OFF			TMR1ON = 0
#define		TIMER2ON		  	TMR2ON = 1
#define		TIMER2OFF		  	TMR2ON = 0
#define		TIMER3ON		  	TMR3ON = 1
#define		TIMER3OFF		  	TMR3ON = 0

#define		rising				1
#define		falling				0
#define		INT0_rising			INTEDG0 = rising
#define		INT0_falling		INTEDG0 = falling

//******************************************************************
//*	EUSART BUFFER DEFINITIONS
//******************************************************************
//1,2,4,8,16,32,64,128 or 256 bytes 
#define EUSART_RX_BUFFER_SIZE  16     
#define EUSART_RX_BUFFER_MASK ( EUSART_RX_BUFFER_SIZE - 1 )
// 1,2,4,8,16,32,64,128 or 256 bytes 
#define EUSART_TX_BUFFER_SIZE  2      
#define EUSART_TX_BUFFER_MASK ( EUSART_TX_BUFFER_SIZE - 1 )
unsigned char EUSART_RxBuf[EUSART_RX_BUFFER_SIZE],EUSART_TxBuf[EUSART_TX_BUFFER_SIZE];
unsigned char EUSART_TxHead,EUSART_TxTail,EUSART_RxHead,EUSART_RxTail;
/*
void cleaner(void) //loads receive buffer with 0x00
{
	unsigned int x;
	for(x=0;x<EUSART_RX_BUFFER_SIZE-1;++x)
		EUSART_RxBuf[x] = 0;
	for(x=0;x<EUSART_TX_BUFFER_SIZE-1;++x)
		EUSART_TxBuf[x] = 0;
  	EUSART_RxTail = 0x00;	//flush receive buffer 
  	EUSART_RxHead = 0x00;
  	EUSART_TxTail = 0x00;	//flush transmit buffer
  	EUSART_TxHead = 0x00;
}

	
void putch(char c)
{
   while(!TRMT) 	// TRMT is set when TSR is empty
      continue;

  TXREG = c;		// load the register
}

unsigned int recvchar(void)
{
  unsigned char tmptail;
                                // wait for incomming usart_data  
  while ( EUSART_RxHead == EUSART_RxTail );
                                // calculate buffer index  
  tmptail = ( EUSART_RxTail + 1 ) & EUSART_RX_BUFFER_MASK;
  EUSART_RxTail = tmptail;        // store new index  

  return EUSART_RxBuf[tmptail];   // return usart_data  
}

unsigned int sendchar(unsigned int usart_data)
{
  unsigned char tmphead;
                                // calculate buffer index 
  tmphead = ( EUSART_TxHead + 1 ) & EUSART_TX_BUFFER_MASK;
                                // wait for free space in buffer 
  while ( tmphead == EUSART_TxTail );
                                // store usart_data in buffer 
  EUSART_TxBuf[tmphead] = (unsigned char)usart_data;
  EUSART_TxHead = tmphead;        // store new index 

  TXIE = 1;            // enable TX interrupt 

  return usart_data;
}

unsigned char CharInQueue(void)
{
  return(EUSART_RxHead != EUSART_RxTail);
}

//******************************************************************
//*	Init EUSART Function
//******************************************************************
void init_EUSART(void)
{
  SPBRG = 25;	//8MHZ = 25 - 4MHZ = 12 FOR 19200BAUD
  TRISC7 = 1;	//receive pin
  TRISC6 = 0;	//transmit pin
  TXSTA = 0x04;	//high speed baud rate	BRGH = 1
  RCSTA = 0x80;	//enable serial port and serial port pins
  
  EUSART_RxTail = 0x00;	//flush receive buffer 
  EUSART_RxHead = 0x00;
  EUSART_TxTail = 0x00;	//flush transmit buffer
  EUSART_TxHead = 0x00;
  RCIP = 1;			//receive interrupt = high priority
  TXIP = 1;			//transmit interrupt = high priority
  RCIE = 1;			//enable receive interrupt
  PEIE = 1;			//enable all unmasked peripheral interrupts
  GIE = 1;			//enable all unmasked interrupts
  CREN = 1;			//enable EUSART1 receiver
  TXIE = 0;			//disable EUSART1 transmit interrupt
  TXEN = 1;			//transmitter enabled
}
*/
void interrupt EUSART_TIMER(void)
{
//SERVICE THE ZVC EVENT
	if((INT0IF && INT0IE))
	{
		disable_GLOBALint;					//disable interrutps
		INT0IF = 0;							//clear INTO interrupt flag
		
		TMR0H = make8(tmr0_trigger_val,hi); //load TIMER0 trigger value
		TMR0L = make8(tmr0_trigger_val,lo); //always load high byte first
		 
		if (flags.int0_edge == falling)		//falling edge on INT0 pin caused interrupt
		{
			flags.int0_edge = rising;		//set flag for rising edge interrupt enabled
			INT0_rising;					//setup INT0 for rising edge interrupt
		}
		else if(flags.int0_edge == rising)	//rising edge on INT0 pin caused interrupt
		{
			flags.int0_edge = falling;		//set flag for falling edge interrupt enabled
			INT0_falling;					//setup INT0 for falling edge interrupt
		}	
		
		if (++refresh_speed_cntr > REFRESH_SPEED_RATE)
		{
			refresh_speed_cntr = 0;
		}
		
		flags.zvc = 1;
		enable_TMR0int;					//enable TIMER0 interrupt
		disable_INT0int;				//disable INT0 interrupt
		enable_GLOBALint;				//enable global interrupt
	}
	
//SERVICE TIMER0
	if(TMR0IF && TMR0IE)
	{
		disable_GLOBALint;				//disable global interrupt
		TMR0IF = 0;						//clear TIMER0 interrupt flag
		
		if(tmr0_trigger_val != 0x0000)
		{
			trigger_triac;				//sink triac gate
			
			TMR3H = 0xFE;				//400uS trigger pulse 
			TMR3L = 0x6F;
			TMR3IF = 0;
			while(TMR3IF == 0);
					
			block_triac					//source triac gate
		}
		disable_TMR0int;				//disable Timer0 interrupt
		enable_INT0int;					//enable INT0 interrupt
		enable_GLOBALint;				//enable global interrupt
	}
/*	
//SERVICE THE EUSART
 if(RCIF)
 {
   usart_data = RCREG;                   // read the received usart_data 
                                // calculate buffer index 
   tmphead = ( EUSART_RxHead + 1 ) & EUSART_RX_BUFFER_MASK;
   EUSART_RxHead = tmphead;        // store new index 

   if ( tmphead == EUSART_RxTail )
   {
     // ERROR! Receive buffer overflow 
		for(temp16=0;temp16<EUSART_RX_BUFFER_SIZE-1;++temp16)
			EUSART_RxBuf[temp16] = 0;
		for(temp16=0;temp16<EUSART_TX_BUFFER_SIZE-1;++temp16)
			EUSART_TxBuf[temp16] = 0;
	  	EUSART_RxTail = 0x00;	//flush receive buffer 
	  	EUSART_RxHead = 0x00;
	  	EUSART_TxTail = 0x00;	//flush transmit buffer
	  	EUSART_TxHead = 0x00;
   }
                                
   EUSART_RxBuf[tmphead] = usart_data;   // store received usart_data in buffer 
  }

 
  if(TRMT)
  {
                                // check if all usart_data is transmitted  
   if ( EUSART_TxHead != EUSART_TxTail )
   {
                                // calculate buffer index  
     tmptail = ( EUSART_TxTail + 1 ) & EUSART_TX_BUFFER_MASK;
     EUSART_TxTail = tmptail;      // store new index  

    TXREG = EUSART_TxBuf[tmptail];  // start transmitiON  
   }
   else
   {
     TXIE = 0;         // disable TX interrupt  
   }
  }
*/
}
//****************************************************************************
//*	CONVERT TRIGGER ANGLE TO TIME
//* THIS ROUTINE CALCULATES THE AMOUNT OF DELAY INTO A HALF CYCLE BEFORE
//* TRIGGERING THE TRIAC GATE.. THIS ROUTINE IS SET FOR 60Hz OPERATION (Ox2E).
//****************************************************************************
unsigned int angle_to_time(char trigger_angle)
{
	unsigned int angle2time_val;
	if(trigger_angle == 0)
	{	
		angle2time_val = 0;
	}
	else
	{
	angle2time_val = 0xFFFF - (trigger_angle * 0x2E); //each degree ~ 46uS
	}
	return angle2time_val;
}
//****************************************************************************
//*	START MOTOR
//* THIS ROUTINE BRINGS THE MOTOR UP TO TRIAC TRIGGER ANGLE SPECIFIED
//* IN reset_angle
//****************************************************************************
void start_mtr(char reset_angle,char zerocrossings)
{	
	char i,j,k;
	
	tmr0_trigger_val = angle_to_time(150);
	
	for(i=150;i>reset_angle;--i)
	{
		for(j=0;j<zerocrossings;++j)
		{
			do{
				++k;
			}while(flags.zvc == 0);
			flags.zvc = 0;
		}
		tmr0_trigger_val = angle_to_time(i);
	}
	flags.mtr_running = 1;
	current_trigger_angle = reset_angle;
	new_trigger_angle = reset_angle;
}
//****************************************************************************
//*	INITIALIZE ROUTINE
//****************************************************************************
void init()
{
	flags.int0_edge = rising; 
	flags.zvc = 0; 
	flags.mtr_running = 0;
	refresh_speed_cntr = 0;
	current_trigger_angle = 0;
	new_trigger_angle = 0;
//****************************************************************************
//*	INITIALIZE CLOCK AND IO PORTS
//****************************************************************************
	OSCCON = 0x70;		//8 MHz internal clock
	PLLEN = 0;			//PLL OFF
	TRISA = 0b11111000;
	TRISB = 0b11111111;
	TRISC = 0b10111111;
	block_triac;
	led = 1;
//******************************************************************
//*	CONFIGURE A2D AND COMPARATORS
//******************************************************************
	ADCON1 = 0b00001111;	//all digital inputs
	ADON = 0;				//A2D OFF
	CMCON = 0x07;			//comparators OFF
//****************************************************************************
//*	INITIALIZE USART
//****************************************************************************
	//init_EUSART();
//******************************************************************
//*	CONFIGURE AND START TIMER0 AS COUNTER
//******************************************************************
	TMR0IP = 1;		//TMR0 interrupt = high priority
	T0CON = 0x00;	//update rate = 1uS	
	TMR0H = 0x00;	
	TMR0L = 0x00;
	TIMER0ON;
//******************************************************************
//*	CONFIGURE AND START TIMER3 AS COUNTER
//******************************************************************
	T3CON = 0x10;	//update rate = 1uS	
	TMR0H = 0x00;	
	TMR0L = 0x00;
	TIMER3ON;
//******************************************************************
//*	CONFIGURE INTERRUPTS
//******************************************************************
	enable_INT0int;
	enable_TMR0int;
	enable_GLOBALint;
}
//****************************************************************************
//*	MAIN ROUTINE
//****************************************************************************
void main()
{
	init();
	
	do{
		//IF OPTORX LED IS OFF, THEN triac_switch IS LOGICALLY HIGH
		//triac_switch BEING LOGICALLY HIGH TURNS MOTOR OFF
		if(triac_switch)
		{
			new_trigger_angle = 0;
			current_trigger_angle = 0;
			flags.mtr_running = 0;
		}
		//SOFT-START MOTOR FROM DEAD STOP AND RAMP UP TO TRIGGER ANGLE 
		// OF 120 DEGREES.. CROSS ZERO 10 TIMES EVERY DEGREE RAMPED UP
		if(flags.mtr_running == 0 && triac_switch == 0)
		{
			start_mtr(120,10);
		}
//****************************************************************************
//*	ESTABLISH NEW TRIAC TRIGGER ANGLE VALUE IN TIMER0
//****************************************************************************
		tmr0_trigger_val = angle_to_time(current_trigger_angle);
	}while(1);
}

/*
// EXAMPLE CODE FOR RS-232 INTERFACE
//
//****************************************************************************
//*	PROCESS INCOMING SPEED BYTE
//* THIS ROUTINE LOOKS FOR A RECEIVED CHARACTER IN THE RECEIVE QUEUE.
//* IF A CHARACTER HAS BEEN RECEIVED, THE CHARACTER IS USED AS THE NEW
//* TIRAC TRIGGER ANGLE.  OTHERWISE, THE MOTOR WILL RAMP UP TO THE SLOWEST
//* SPEED.
//****************************************************************************
		if(CharInQueue())
		{
			new_trigger_angle = recvchar();
			
			if(new_trigger_angle == 0)
			{
				block_triac;
				new_trigger_angle = 0;
				current_trigger_angle = 0;
				flags.mtr_running = 0;
			}
			else 
			{
				enable_INT0int;
				start_mtr(140,10);
			}

		}
		
//****************************************************************************
//*	SMOOTHLY MOVE TO NEW SPEED
//****************************************************************************
		if(refresh_speed_cntr == REFRESH_SPEED_RATE)
		{
			refresh_speed_cntr = 0;
			
			if (new_trigger_angle > current_trigger_angle)
			{
				++current_trigger_angle;
			}
			else if(new_trigger_angle < current_trigger_angle)
			{
				--current_trigger_angle;
			}
		}
*/